@@ -1,8 +1,6 @@
 # Various tests should have random accuracy draws, probably with lots of
 # weight on 1
 
-library(R.utils)
-
 describe("discover", {
   expect_equiv_deps <- function(deps1, deps2) {
     expect_setequal(attrs_order(deps1), attrs_order(deps2))
@@ -71,7 +69,7 @@ describe("discover", {
   }
   terminates_then <- function(fn, accuracy, ...) {
     function(df) {
-      res <- withTimeout(discover(df, accuracy, ...), timeout = 5, onTimeout = "silent")
+      res <- R.utils::withTimeout(discover(df, accuracy = accuracy, ...), timeout = 5, onTimeout = "silent")
       if (is.null(res)) {
         return(fail("discover() timed out"))
       }
@@ -81,23 +79,23 @@ describe("discover", {
   }
   both_terminate_then <- function(fn, accuracy, ...) {
     function(df1, df2) {
-      res1 <- withTimeout(discover(df1, accuracy, ...), timeout = 5, onTimeout = "silent")
+      res1 <- R.utils::withTimeout(discover(df1, accuracy = accuracy, ...), timeout = 5, onTimeout = "silent")
       expect_true(!is.null(res1))
-      res2 <- withTimeout(discover(df2, accuracy, ...), timeout = 5, onTimeout = "silent")
+      res2 <- R.utils::withTimeout(discover(df2, accuracy = accuracy, ...), timeout = 5, onTimeout = "silent")
       expect_true(!is.null(res2))
       fn(res1, res2)
     }
   }
   terminates_with_and_without_full_cache_then <- function(fn, accuracy, ...) {
     function(df) {
-      res_cache <- withTimeout(
-        discover(df, accuracy, full_cache = TRUE, ...),
+      res_cache <- R.utils::withTimeout(
+        discover(df, accuracy = accuracy, full_cache = TRUE, ...),
         timeout = 5,
         onTimeout = "silent"
       )
       expect_true(!is.null(res_cache))
-      res_nocache <- withTimeout(
-        discover(df, accuracy, full_cache = FALSE, ...),
+      res_nocache <- R.utils::withTimeout(
+        discover(df, accuracy = accuracy, full_cache = FALSE, ...),
         timeout = 5,
         onTimeout = "silent"
       )
@@ -107,14 +105,14 @@ describe("discover", {
   }
   terminates_with_and_without_store_cache_then <- function(fn, accuracy, ...) {
     function(df) {
-      res_store <- withTimeout(
-        discover(df, accuracy, full_cache = TRUE, store_cache = TRUE, ...),
+      res_store <- R.utils::withTimeout(
+        discover(df, accuracy = accuracy, full_cache = TRUE, store_cache = TRUE, ...),
         timeout = 5,
         onTimeout = "silent"
       )
       expect_true(!is.null(res_store))
-      res_nostore <- withTimeout(
-        discover(df, accuracy, full_cache = TRUE, store_cache = FALSE, ...),
+      res_nostore <- R.utils::withTimeout(
+        discover(df, accuracy = accuracy, full_cache = TRUE, store_cache = FALSE, ...),
         timeout = 5,
         onTimeout = "silent"
       )
@@ -124,15 +122,15 @@ describe("discover", {
   }
   terminates_with_and_without_bijection_skip_then <- function(fn, accuracy, ...) {
     function(df) {
-      res_skip <- withTimeout(
-        discover(df, accuracy, full_cache = TRUE, store_cache = TRUE, skip_bijections = TRUE, ...),
+      res_skip <- R.utils::withTimeout(
+        discover(df, accuracy = accuracy, full_cache = TRUE, store_cache = TRUE, skip_bijections = TRUE, ...),
         timeout = 5,
         onTimeout = "silent"
       )
       if (is.null(res_skip))
         return(fail("discover() with bijection skip timed out"))
-      res_noskip <- withTimeout(
-        discover(df, accuracy, full_cache = TRUE, store_cache = TRUE, skip_bijections = FALSE, ...),
+      res_noskip <- R.utils::withTimeout(
+        discover(df, accuracy = accuracy, full_cache = TRUE, store_cache = TRUE, skip_bijections = FALSE, ...),
         timeout = 5,
         onTimeout = "silent"
       )
@@ -150,7 +148,15 @@ describe("discover", {
     }
     forall(
       gen_df(4, 6),
-      two_copies(both_terminate_then(expect_equiv_deps, accuracy = 1))
+      two_copies(both_terminate_then(expect_equiv_deps, accuracy = 1, method = "DFD"))
+    )
+    forall(
+      gen_df(4, 6),
+      two_copies(both_terminate_then(expect_equiv_deps, accuracy = 1, method = "FDHitsSep"))
+    )
+    forall(
+      gen_df(4, 6),
+      two_copies(both_terminate_then(expect_equiv_deps, accuracy = 1, method = "FDHitsJoint"))
     )
   })
   it("returns dependencies where shared dependant <=> not sub/supersets for determinants", {
@@ -170,7 +176,9 @@ describe("discover", {
         }
       }
     }
-    forall(gen_df(4, 6), terminates_then(has_non_nested_determinant_sets, 1))
+    forall(gen_df(4, 6), terminates_then(has_non_nested_determinant_sets, 1, method = "DFD"))
+    forall(gen_df(4, 6), terminates_then(has_non_nested_determinant_sets, 1, method = "FDHitsSep"))
+    forall(gen_df(4, 6), terminates_then(has_non_nested_determinant_sets, 1, method = "FDHitsJoint"))
   })
   it("is invariant to an attribute's values being permuted", {
     gen_perm <- function(vals) {
@@ -198,7 +206,17 @@ describe("discover", {
     }
     forall(
       gen_df_and_value_perm(4, 6),
-      both_terminate_then(expect_equiv_deps, 1),
+      both_terminate_then(expect_equiv_deps, 1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_value_perm(4, 6),
+      both_terminate_then(expect_equiv_deps, 1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_value_perm(4, 6),
+      both_terminate_then(expect_equiv_deps, 1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
@@ -233,7 +251,17 @@ describe("discover", {
     }
     forall(
       gen_df_and_type_change(4, 6),
-      both_terminate_then(expect_equiv_deps_except_classes, 1),
+      both_terminate_then(expect_equiv_deps_except_classes, 1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_type_change(4, 6),
+      both_terminate_then(expect_equiv_deps_except_classes, 1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_type_change(4, 6),
+      both_terminate_then(expect_equiv_deps_except_classes, 1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
@@ -247,12 +275,19 @@ describe("discover", {
       sep = c(TRUE, TRUE, NA, NA)
     )
     df2 <- df1[, c("l", "j", "t", "b", "u", "sep")]
-    terminates_with_and_without_cache <- terminates_with_and_without_full_cache_then(
-      \(x, y) {},
-      1
-    )
-    terminates_with_and_without_cache(df1)
-    terminates_with_and_without_cache(df2)
+    terminates_with_and_without_cache <- function(...) {
+      terminates_with_and_without_full_cache_then(
+        \(x, y) {},
+        1,
+        ...
+      )
+    }
+    terminates_with_and_without_cache(method = "DFD")(df1)
+    terminates_with_and_without_cache(method = "FDHitsSep")(df1)
+    terminates_with_and_without_cache(method = "FDHitsJoint")(df1)
+    terminates_with_and_without_cache(method = "DFD")(df2)
+    terminates_with_and_without_cache(method = "FDHitsSep")(df2)
+    terminates_with_and_without_cache(method = "FDHitsJoint")(df2)
   })
   it("is invariant, under reordering, to attributes being reordered", {
     gen_df_and_attr_perm <- function(
@@ -271,7 +306,17 @@ describe("discover", {
 
     forall(
       gen_df_and_attr_perm(4, 6),
-      both_terminate_then(expect_equiv_deps, 1),
+      both_terminate_then(expect_equiv_deps, 1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_attr_perm(4, 6),
+      both_terminate_then(expect_equiv_deps, 1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_attr_perm(4, 6),
+      both_terminate_then(expect_equiv_deps, 1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
@@ -287,11 +332,21 @@ describe("discover", {
     }
     forall(
       gen_df_and_remove_col(4, 6),
-      both_terminate_then(expect_equiv_non_removed_attr_deps, 1),
+      both_terminate_then(expect_equiv_non_removed_attr_deps, 1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_remove_col(4, 6),
+      both_terminate_then(expect_equiv_non_removed_attr_deps, 1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_remove_col(4, 6),
+      both_terminate_then(expect_equiv_non_removed_attr_deps, 1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
-  it("is invariant to changes of accuracy within same required row count", {
+  it("is invariant to changes of accuracy within same required row count when using DFD", {
     gen_df_and_accuracy_nrow <- function(nrow, ncol, remove_dup_rows = FALSE) {
       gen_df(nrow, ncol, minrow = 1L, mincol = 1L, remove_dup_rows) |>
         gen.and_then(\(df) list(df, gen.int(nrow(df)))) |>
@@ -305,16 +360,16 @@ describe("discover", {
     }
     both_bounds_terminate_then <- function(fn, ...) {
       function(df, low, high) {
-        res1 <- withTimeout(discover(df, low, ...), timeout = 5, onTimeout = "silent")
+        res1 <- R.utils::withTimeout(discover(df, accuracy = low, ...), timeout = 5, onTimeout = "silent")
         expect_true(!is.null(res1))
-        res2 <- withTimeout(discover(df, high, ...), timeout = 5, onTimeout = "silent")
+        res2 <- R.utils::withTimeout(discover(df, accuracy = high, ...), timeout = 5, onTimeout = "silent")
         expect_true(!is.null(res2))
         fn(res1, res2)
       }
     }
     forall(
       gen_df_and_accuracy_nrow(4, 6),
-      both_bounds_terminate_then(expect_equiv_deps),
+      both_bounds_terminate_then(expect_equiv_deps, method = "DFD"),
       curry = TRUE
     )
   })
@@ -330,7 +385,17 @@ describe("discover", {
     }
     forall(
       gen_df_and_remove_row(4, 6),
-      both_terminate_then(expect_det_subsets_kept, 1),
+      both_terminate_then(expect_det_subsets_kept, 1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_remove_row(4, 6),
+      both_terminate_then(expect_det_subsets_kept, 1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_remove_row(4, 6),
+      both_terminate_then(expect_det_subsets_kept, 1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
@@ -346,7 +411,17 @@ describe("discover", {
     }
     forall(
       gen_df_and_name_change(4, 6),
-      both_terminate_then(expect_equiv_deps_except_names, 1),
+      both_terminate_then(expect_equiv_deps_except_names, 1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_name_change(4, 6),
+      both_terminate_then(expect_equiv_deps_except_names, 1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_name_change(4, 6),
+      both_terminate_then(expect_equiv_deps_except_names, 1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
@@ -366,8 +441,12 @@ describe("discover", {
       )
     )
     stopifnot(df[1, "time"] != df[2, "time"])
-    deps <- discover(df, 1)
+    deps <- discover(df, method = "DFD")
     expect_length(deps[dependant(deps) == "time"], 0L)
+    deps2 <- discover(df, method = "FDHitsSep")
+    expect_length(deps2[dependant(deps2) == "time"], 0L)
+    deps3 <- discover(df, method = "FDHitsJoint")
+    expect_length(deps3[dependant(deps3) == "time"], 0L)
   })
   it("correctly simplifies floating-point numbers to high accuracy", {
     df <- data.frame(
@@ -381,14 +460,42 @@ describe("discover", {
       )
     )
     expect_identical(
-      discover(df, 1, digits = 8),
+      discover(df, digits = 8, method = "DFD"),
       functional_dependency(
         list(list(character(), "x"), list(character(), "y")),
         c("x", "y")
       )
     )
     expect_identical(
-      discover(df, 1, digits = 15),
+      discover(df, digits = 8, method = "FDHitsSep"),
+      functional_dependency(
+        list(list(character(), "x"), list(character(), "y")),
+        c("x", "y")
+      )
+    )
+    expect_identical(
+      discover(df, digits = 8, method = "FDHitsJoint"),
+      functional_dependency(
+        list(list(character(), "x"), list(character(), "y")),
+        c("x", "y")
+      )
+    )
+    expect_identical(
+      discover(df, digits = 15, method = "DFD"),
+      functional_dependency(
+        list(list(character(), "x")),
+        c("x", "y")
+      )
+    )
+    expect_identical(
+      discover(df, digits = 15, method = "FDHitsSep"),
+      functional_dependency(
+        list(list(character(), "x")),
+        c("x", "y")
+      )
+    )
+    expect_identical(
+      discover(df, digits = 15, method = "FDHitsJoint"),
       functional_dependency(
         list(list(character(), "x")),
         c("x", "y")
@@ -402,8 +509,8 @@ describe("discover", {
     }
     terminates_with_exclusion_then_no_trival <- function(accuracy, ...) {
       function(df, attr) {
-        deps <- withTimeout(
-          discover(df, accuracy, exclude = attr, ...),
+        deps <- R.utils::withTimeout(
+          discover(df, accuracy = accuracy, exclude = attr, ...),
           timeout = 5,
           onTimeout = "silent"
         )
@@ -416,7 +523,17 @@ describe("discover", {
     }
     forall(
       gen_df_and_exclude(4, 6),
-      terminates_with_exclusion_then_no_trival(1),
+      terminates_with_exclusion_then_no_trival(1, method = "DFD"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_exclude(4, 6),
+      terminates_with_exclusion_then_no_trival(1, method = "FDHitsSep"),
+      curry = TRUE
+    )
+    forall(
+      gen_df_and_exclude(4, 6),
+      terminates_with_exclusion_then_no_trival(1, method = "FDHitsJoint"),
       curry = TRUE
     )
   })
@@ -424,15 +541,15 @@ describe("discover", {
     exclude_and_exclude_class_terminate_then <- function(fn, accuracy, class, ...) {
       function(df) {
         attrs_with_class <- names(df)[vapply(df, inherits, logical(1), class)]
-        deps1 <- withTimeout(
-          discover(df, accuracy, exclude = attrs_with_class, ...),
+        deps1 <- R.utils::withTimeout(
+          discover(df, accuracy = accuracy, exclude = attrs_with_class, ...),
           timeout = 5,
           onTimeout = "silent"
         )
         if (is.null(deps1))
           return(fail("discover() with exclude timed out"))
-        deps2 <- withTimeout(
-          discover(df, accuracy, exclude_class = class, ...),
+        deps2 <- R.utils::withTimeout(
+          discover(df, accuracy = accuracy, exclude_class = class, ...),
           timeout = 5,
           onTimeout = "silent"
         )
@@ -446,7 +563,26 @@ describe("discover", {
       exclude_and_exclude_class_terminate_then(
         expect_equiv_deps,
         accuracy = 1,
-        "logical"
+        "logical",
+        method = "DFD"
+      )
+    )
+    forall(
+      gen_df(4, 6),
+      exclude_and_exclude_class_terminate_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        "logical",
+        method = "FDHitsSep"
+      )
+    )
+    forall(
+      gen_df(4, 6),
+      exclude_and_exclude_class_terminate_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        "logical",
+        method = "FDHitsJoint"
       )
     )
   })
@@ -460,22 +596,25 @@ describe("discover", {
     same_under_constraints_and_filtering <- function(
       x,
       dependants,
-      detset_limit
+      detset_limit,
+      method,
+      ...
     ) {
-      by_arguments <- withTimeout(
+      by_arguments <- R.utils::withTimeout(
         discover(
           x,
-          1,
           dependants = dependants,
-          detset_limit = detset_limit
+          detset_limit = detset_limit,
+          method = method,
+          ...
         ),
         timeout = 5,
         onTimeout = "silent"
       )
       if (is.null(by_arguments))
         return(fail("discover() with filtering arguments timed out"))
-      by_filtering <- withTimeout(
-        discover(x, 1),
+      by_filtering <- R.utils::withTimeout(
+        discover(x, method = method, ...),
         timeout = 5,
         onTimeout = "silent"
       )
@@ -493,7 +632,8 @@ describe("discover", {
           list(
             gen.pure(x),
             gen.sample_resampleable(names(x), from = 0, to = ncol(x)),
-            gen.element(0:(ncol(x) - 1L))
+            gen.element(0:(ncol(x) - 1L)),
+            gen.element(c("DFD", "FDHitsSep", "FDHitsJoint"))
           )
         }),
       same_under_constraints_and_filtering,
@@ -508,14 +648,27 @@ describe("discover", {
       d = c(1L, 1L, NA, 1L),
       e = c(NA, 1L, 1L, 0L)
     )
-    same_under_constraints_and_filtering(x, c("a", "c"), 2L)
+    same_under_constraints_and_filtering(x, c("a", "c"), 2L, method = "DFD")
+    same_under_constraints_and_filtering(x, c("a", "c"), 2L, method = "FDHitsSep")
+    same_under_constraints_and_filtering(x, c("a", "c"), 2L, method = "FDHitsJoint")
   })
   it("gives dependencies for unique attributes (in case don't want them as key)", {
     df <- data.frame(A = 1:3, B = c(1, 1, 2), C = c(1, 2, 2))
-    deps <- discover(df, 1)
-    A_deps <- dependant(deps) == "A"
-    A_detsets <- detset(deps[A_deps])
-    expect_identical(A_detsets, list(c("B", "C")))
+
+    deps_dfd <- discover(df, method = "DFD")
+    A_deps_dfd <- dependant(deps_dfd) == "A"
+    A_detsets_dfd <- detset(deps_dfd[A_deps_dfd])
+    expect_identical(A_detsets_dfd, list(c("B", "C")))
+
+    deps_sep <- discover(df, method = "FDHitsSep")
+    A_deps_sep <- dependant(deps_sep) == "A"
+    A_detsets_sep <- detset(deps_sep[A_deps_sep])
+    expect_identical(A_detsets_sep, list(c("B", "C")))
+
+    deps_jnt <- discover(df, method = "FDHitsJoint")
+    A_deps_jnt <- dependant(deps_jnt) == "A"
+    A_detsets_jnt <- detset(deps_jnt[A_deps_jnt])
+    expect_identical(A_detsets_jnt, list(c("B", "C")))
   })
   it("finds dependencies for the team data in test-synthesise", {
     df <- data.frame(
@@ -540,7 +693,6 @@ describe("discover", {
         'HI', 'HI', 'MA', 'MA', 'TX'
       )
     )
-    deps <- discover(df, 1)
     expected_deps <- functional_dependency(
       list(
         list(c('player_name', 'jersey_num'), "team"),
@@ -556,8 +708,17 @@ describe("discover", {
       c("team", "jersey_num", "player_name", "city", "state")
     )
 
-    expect_identical(attrs_order(deps), attrs_order(expected_deps))
-    expect_true(all(is.element(expected_deps, deps)))
+    deps_dfd <- discover(df, method = "DFD")
+    expect_identical(attrs_order(deps_dfd), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_dfd)))
+
+    deps_sep <- discover(df, method = "FDHitsSep")
+    expect_identical(attrs_order(deps_sep), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_sep)))
+
+    deps_jnt <- discover(df, method = "FDHitsJoint")
+    expect_identical(attrs_order(deps_jnt), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_jnt)))
   })
   it("finds dependencies for the team data in original's edit demo", {
     df <- data.frame(
@@ -566,7 +727,6 @@ describe("discover", {
       state = c("MA", "IL", "FL", "TX", "HI", "TX"),
       roster_size = c(20L, 21L, 20L, 20L, 19L, 21L)
     )
-    deps <- discover(df, 1)
     expected_deps <- functional_dependency(
       list(
         list("city", "team"),
@@ -578,8 +738,18 @@ describe("discover", {
       ),
       c("team", "city", "state", "roster_size")
     )
-    expect_identical(attrs_order(deps), attrs_order(expected_deps))
-    expect_true(all(is.element(expected_deps, deps)))
+
+    deps_dfd <- discover(df, method = "DFD")
+    expect_identical(attrs_order(deps_dfd), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_dfd)))
+
+    deps_sep <- discover(df, method = "FDHitsSep")
+    expect_identical(attrs_order(deps_sep), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_sep)))
+
+    deps_jnt <- discover(df, method = "FDHitsJoint")
+    expect_identical(attrs_order(deps_jnt), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_jnt)))
   })
   it("finds dependencies for Wikipedia 1NF->2NF->3NF example", {
     df <- data.frame(
@@ -599,7 +769,6 @@ describe("discover", {
       Genre_Name = rep(c("Tutorial", "Popular science"), each = 2),
       Publisher_ID = rep(1:2, each = 2)
     )
-    deps <- discover(df, 1)
     expected_deps <- functional_dependency(
       list(
         list("Title", "Author"),
@@ -622,42 +791,114 @@ describe("discover", {
         "Publisher_ID"
       )
     )
-    expect_identical(attrs_order(deps), attrs_order(expected_deps))
-    expect_true(all(is.element(expected_deps, deps)))
+    deps_dfd <- discover(df, method = "DFD")
+    deps_sep <- discover(df, method = "FDHitsSep")
+    deps_jnt <- discover(df, method = "FDHitsJoint")
+    expect_identical(attrs_order(deps_dfd), attrs_order(expected_deps))
+    expect_identical(attrs_order(deps_sep), attrs_order(expected_deps))
+    expect_identical(attrs_order(deps_jnt), attrs_order(expected_deps))
+    expect_true(all(is.element(expected_deps, deps_dfd)))
+    expect_true(all(is.element(expected_deps, deps_sep)))
+    expect_true(all(is.element(expected_deps, deps_jnt)))
   })
   it("correctly handles attributes with non-df-standard names", {
     df <- data.frame(1:3, c(1, 1, 2), c(1, 2, 2)) |>
       stats::setNames(c("A 1", "B 2", "C 3"))
-    deps <- discover(df, 1)
-    A_1_deps <- dependant(deps) == "A 1"
-    A_1_detsets <- detset(deps[A_1_deps])
-    expect_identical(A_1_detsets, list(c("B 2", "C 3")))
+
+    deps_dfd <- discover(df, method = "DFD")
+    A_1_deps_dfd <- dependant(deps_dfd) == "A 1"
+    A_1_detsets_dfd <- detset(deps_dfd[A_1_deps_dfd])
+    expect_identical(A_1_detsets_dfd, list(c("B 2", "C 3")))
+
+    deps_sep <- discover(df, method = "FDHitsSep")
+    A_1_deps_sep <- dependant(deps_sep) == "A 1"
+    A_1_detsets_sep <- detset(deps_sep[A_1_deps_sep])
+    expect_identical(A_1_detsets_sep, list(c("B 2", "C 3")))
+
+    deps_jnt <- discover(df, method = "FDHitsJoint")
+    A_1_deps_jnt <- dependant(deps_jnt) == "A 1"
+    A_1_detsets_jnt <- detset(deps_jnt[A_1_deps_jnt])
+    expect_identical(A_1_detsets_jnt, list(c("B 2", "C 3")))
   })
   it("expects attribute names to be unique", {
     df <- data.frame(A = 1:3, B = c(1, 1, 2), A = c(1, 2, 2), check.names = FALSE)
-    expect_error(discover(df, 1), "^duplicate column names: A$")
+    expect_error(discover(df, method = "DFD"), "^duplicate column names: A$")
+    expect_error(discover(df, method = "FDHitsSep"), "^duplicate column names: A$")
+    expect_error(discover(df, method = "FDHitsJoint"), "^duplicate column names: A$")
   })
   it("gets the same results with and without storing partitions", {
     forall(
       gen_df(20, 5),
-      terminates_with_and_without_full_cache_then(expect_equiv_deps, accuracy = 1),
+      terminates_with_and_without_full_cache_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        method = "DFD"
+      ),
       shrink.limit = Inf
     )
     forall(
       gen_df(20, 5),
-      terminates_with_and_without_full_cache_then(expect_equiv_deps, accuracy = 3/4),
+      terminates_with_and_without_full_cache_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        method = "FDHitsSep"
+      ),
+      shrink.limit = Inf
+    )
+    forall(
+      gen_df(20, 5),
+      terminates_with_and_without_full_cache_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        method = "FDHitsJoint"
+      ),
+      shrink.limit = Inf
+    )
+    forall(
+      gen_df(20, 5),
+      terminates_with_and_without_full_cache_then(
+        expect_equiv_deps,
+        accuracy = 3/4,
+        method = "DFD"
+      ),
       shrink.limit = Inf
     )
   })
-  it("is invariant to whether partition is transferred between dependants", {
+  it("is invariant to whether partition is transferred between dependants, also for accuracy < 1 for DFD", {
     forall(
       gen_df(20, 5),
-      terminates_with_and_without_store_cache_then(expect_equiv_deps, accuracy = 1),
+      terminates_with_and_without_store_cache_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        method = "DFD"
+      ),
       shrink.limit = Inf
     )
     forall(
       gen_df(20, 5),
-      terminates_with_and_without_store_cache_then(expect_equiv_deps, accuracy = 3/4),
+      terminates_with_and_without_store_cache_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        method = "FDHitsSep"
+      ),
+      shrink.limit = Inf
+    )
+    forall(
+      gen_df(20, 5),
+      terminates_with_and_without_store_cache_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        method = "FDHitsJoint"
+      ),
+      shrink.limit = Inf
+    )
+    forall(
+      gen_df(20, 5),
+      terminates_with_and_without_store_cache_then(
+        expect_equiv_deps,
+        accuracy = 3/4,
+        method = "DFD"
+      ),
       shrink.limit = Inf
     )
   })
@@ -683,26 +924,47 @@ describe("discover", {
       d = c(NA, NA, TRUE),
       e = c(FALSE, TRUE, NA)
     )
-    invariant_to_bijection_skip <- terminates_with_and_without_bijection_skip_then(
-      expect_equiv_deps,
-      accuracy = 1
-    )
-    invariant_to_bijection_skip(df1)
-    invariant_to_bijection_skip(df2)
-    invariant_to_bijection_skip(df3)
+    invariant_to_bijection_skip <- function(...) {
+      terminates_with_and_without_bijection_skip_then(
+        expect_equiv_deps,
+        accuracy = 1,
+        ...
+      )
+    }
+    invariant_to_bijection_skip(method = "DFD")(df1)
+    invariant_to_bijection_skip(method = "DFD")(df2)
+    invariant_to_bijection_skip(method = "DFD")(df3)
+    invariant_to_bijection_skip(method = "FDHitsSep")(df1)
+    invariant_to_bijection_skip(method = "FDHitsSep")(df2)
+    invariant_to_bijection_skip(method = "FDHitsSep")(df3)
+    invariant_to_bijection_skip(method = "FDHitsJoint")(df1)
+    invariant_to_bijection_skip(method = "FDHitsJoint")(df2)
+    invariant_to_bijection_skip(method = "FDHitsJoint")(df3)
     forall(
       gen_df(20, 5),
-      invariant_to_bijection_skip,
+      invariant_to_bijection_skip(method = "DFD"),
+      shrink.limit = Inf
+    )
+    forall(
+      gen_df(20, 5),
+      invariant_to_bijection_skip(method = "FDHitsSep"),
+      shrink.limit = Inf
+    )
+    forall(
+      gen_df(20, 5),
+      invariant_to_bijection_skip(method = "FDHitsJoint"),
       shrink.limit = Inf
     )
   })
   it(
     paste(
-      "is invariant to:",
+      "for accuracy = 1, is invariant to:",
+      "- having a non-false keep_rownames vs. adding row names as first column",
+      "- method used",
       "- excluding a class vs. excluding attributes in that class",
       "- filtering by arguments (dependants/detset_limit) or by subsetting results",
       "- whether partition is transferred between dependants",
-      "- whether bijections are skipped (accuracy = 1)",
+      "- whether bijections are skipped",
       sep = "\n"
     ),
     {
@@ -711,13 +973,32 @@ describe("discover", {
           gen.and_then(\(x) {
             list(
               gen.pure(x),
+              gen.choice(gen.element(c(FALSE, TRUE)), gen_attr_name(9)),
               gen.sample_resampleable(names(x), from = 0, to = ncol(x)),
               gen.element(0:(ncol(x) - 1L))
             )
           }),
-        function(df, dependants, detset_limit) {
+        function(df, keep_rownames, dependants, detset_limit) {
           arglists <- expand.grid(
-            list(list(df = df, accuracy = 1)),
+            if (!isFALSE(keep_rownames)) {
+              tmp <- if (isTRUE(keep_rownames)) "row" else keep_rownames
+              nms <- make.unique(c(names(df), tmp))
+              nm <- nms[[length(nms)]]
+              list(
+                list(df = df, accuracy = 1, keep_rownames = keep_rownames),
+                list(
+                  df = cbind(setNames(data.frame(rownames(df)), nm), df),
+                  accuracy = 1,
+                  keep_rownames = FALSE
+                )
+              )
+            }else
+              list(list(df = df, accuracy = 1, keep_rownames = FALSE)),
+            list(
+              list(method = "DFD"),
+              list(method = "FDHitsSep"),
+              list(method = "FDHitsJoint")
+            ),
             list(
               list(exclude_class = "logical"),
               list(exclude = names(df)[vapply(df, is.logical, logical(1))])
@@ -742,7 +1023,7 @@ describe("discover", {
           results <- lapply(
             arglists,
             \(lst) {
-              base <- withTimeout(
+              base <- R.utils::withTimeout(
                 do.call(discover, lst),
                 timeout = 5,
                 onTimeout = "silent"
@@ -767,23 +1048,33 @@ describe("discover", {
   it("returns a minimal functional dependency set", {
     forall(
       gen_df(6, 7),
-      terminates_then(is_valid_minimal_functional_dependency, 1)
+      terminates_then(is_valid_minimal_functional_dependency, 1, method = "DFD")
     )
-  })
-})
-
-describe("generate_next_seeds", {
-  it("generates all single attributes if min_deps and max_non_deps are empty", {
     forall(
-      gen.int(10),
-      \(n) {
-        powerset <- nonempty_powerset(n, use_visited = FALSE)
-        lhs_attr_nodes <- to_nodes(seq_len(n), powerset)
-        expect_setequal(
-          generate_next_seeds(integer(), integer(), lhs_attr_nodes, powerset, detset_limit = n),
-          lhs_attr_nodes
-        )
-      }
+      gen_df(6, 7),
+      terminates_then(is_valid_minimal_functional_dependency, 1, method = "FDHitsSep")
+    )
+    forall(
+      gen_df(6, 7),
+      terminates_then(is_valid_minimal_functional_dependency, 1, method = "FDHitsJoint")
+    )
+  })
+  it("can include the rownames, equivalent to adding them as first column", {
+    x <- data.frame(
+      a = c(1, 1, 1, 2, 2, 3, 3, 3, 4),
+      b = c(1, 1, 1, 1, 1, 2, 2, 2, 3),
+      row.names = letters[1:9]
+    )
+    default <- discover(x, keep_rownames = TRUE)
+    explicit <- discover(x, keep_rownames = "id")
+    expected <- discover(x)
+    expect_setequal(
+      default,
+      discover(cbind(data.frame(row = rownames(x)), x))
+    )
+    expect_setequal(
+      explicit,
+      discover(cbind(data.frame(id = rownames(x)), x))
     )
   })
 })
